import os
import sys
from typing import Dict
from dataclasses import dataclass, field

from bgnlp.tools.mappings import POS_ID2LABEL, POS_LABEL2ID


PACKAGE_DIR = os.path.dirname(sys.modules["bgnlp"].__file__)


@dataclass
class ModelConfig:
    model_path: str
    device: str = "cpu"


@dataclass
class PosTaggerConfig(ModelConfig):
    device: str = "cpu"
    model_url: str = "https://drive.google.com/uc?id=1MHADXPM5oYRMz4nyPqaKeW2zQSctOxBQ"
    model_path: str = os.path.join(
        PACKAGE_DIR, "serialized", "models", "pos-bert.pt"
    )

    base_model_id: str = "rmihaylov/bert-base-bg"
    label2id: Dict[str, int] = field(default_factory=lambda: POS_LABEL2ID)
    id2label: Dict[int, str] = field(default_factory=lambda: POS_ID2LABEL)


@dataclass
class LemmaTaggerConfig(ModelConfig):
    device: str = "cpu"
    model_url: str = "https://drive.google.com/uc?id=1U0Tb1AN0Rzdz_WsjMayEKK1rHuyZWdfK"
    model_path: str = os.path.join(
        PACKAGE_DIR, "serialized", "models", "lemma-bert.pt"
    )

    vocab_path: str = os.path.join(
        PACKAGE_DIR, "serialized", "vocabs", "cb-vocab.pt"
    )


@dataclass
class NerTaggerConfig(ModelConfig):
    device: str = "cpu"
    # Since I have uploaded this model to Huggingface's Model Hub, it is being 
    # downloaded only using this model ID:
    model_path: str = "auhide/bert-bg-ner"



import logging
import random

import torch
from torch import nn
from torchtext.vocab import Vocab
from torch.utils.data import DataLoader, Dataset


logging.basicConfig(
    format="%(asctime)s|%(levelname)s| %(message)s", 
    level=logging.INFO
)


class Seq2Seq(nn.Module):
    
    def __init__(
        self, 
        encoder: torch.nn.Module, 
        decoder: torch.nn.Module, 
        reverse_input=True
    ):
        super().__init__()
        
        self.encoder = encoder
        self.decoder = decoder

        self.reverse_input = reverse_input

    def forward(self, x: torch.Tensor, y: torch.Tensor, teacher_forcing_ratio=0.5):
        batch_size, tgt_len = y.shape[0], y.shape[1]
        vocab_size = self.decoder.vocab_size

        # Tensor in which we are going to store the Decoder outputs.
        outputs = torch.zeros(size=(batch_size, tgt_len, vocab_size))

        # Reversing the sequences in the input tensor.
        # In the paper it's stated that it is extremely valuable, and makes a difference.
        if self.reverse_input:
            x = torch.flip(x, [1])
        
        enc_out, hidden = self.encoder(x)

        # The start token:
        curr_input = y[:, 0]

        for t in range(1, tgt_len):
            out, hidden = self.decoder(curr_input, hidden, enc_out)
            outputs[:, t, :] = out

            teacher_force = teacher_forcing_ratio > random.random()

            # Getting the top prediction.
            top_pred = out.argmax(1)

            # If we are in a teacher forcing state - use actual next token
            # as an input to the Decoder, else - use the top prediction.
            curr_input = y[:, t] if teacher_force else top_pred

        return outputs


class Encoder(nn.Module):

    def __init__(
        self, 
        vocab_size: int, 
        embed_size: int, 
        hidden_size: int, 
        padding_idx: int, 
        bidirectional=True, 
        dropout=0.1
    ):
        super().__init__()

        self.hidden_size = hidden_size
        self.num_layers = 1
        self.bidirectional = bidirectional

        self.embed = nn.Embedding(vocab_size, embed_size)
        self.gru = nn.GRU(
            input_size=embed_size, 
            hidden_size=hidden_size,
            num_layers=self.num_layers,
            batch_first=True,
            bidirectional=bidirectional
        )
        self.dropout = nn.Dropout(dropout)

        directions = 2 if self.bidirectional else 1
        
        # A fully-connected layer used for the generation of a single hidden state.
        # It's used since we have twice as many hidden states (we are using a BiLSTM).
        self.fc = nn.Linear(directions * hidden_size, hidden_size)

    def forward(self, x: torch.Tensor):
        embedded_x = self.dropout(self.embed(x))
        out, hidden = self.gru(embedded_x)
        # out.shape: (BATCH_SIZE, SEQ_LEN, HID_SIZE)

        # Concatenating the hidden states in both directions along the 2nd dimension.
        hidden = torch.tanh(
            self.fc(
                torch.cat(
                    (hidden[-2, :, :], hidden[-1, :, :]), 
                    dim=1
                )
            )
        )
        # hidden.shape: (BATCH_SIZE, HIDDEN_DIM)

        return out, hidden


class Attention(nn.Module):
    
    def __init__(self, encoder_hidden_dim: int, decoder_hidden_dim: int):
        super().__init__()
        
        self.attn = nn.Linear((encoder_hidden_dim * 2) + decoder_hidden_dim, decoder_hidden_dim)
        self.v = nn.Linear(decoder_hidden_dim, 1, bias=False)
        
    def forward(self, hidden, encoder_outputs):
        # hidden.shape: (BATCH_SIZE, DEC_HID_DIM)
        # encoder_outputs.shape: (BATCH_SIZE, SEQ_LEN, ENC_HID_DIM * 2)
        seq_len = encoder_outputs.shape[1]

        # Repeating the previous decoder hidden state 'seq_len' times, since
        # the decoder's hidden state doesn't have a SEQ_LEN dimension.
        hidden = hidden.unsqueeze(1).repeat(1, seq_len, 1)
        # After the change hidden.shape will be (BATCH_SIZE, SEQ_LEN, DEC_HID_DIM)

        # Calculating the energy between the prev. decoder hidden state and 
        # the hidden state of the encoder.
        energy = torch.sigmoid(
            self.attn(
                torch.concat((hidden, encoder_outputs), dim=-1)
            )
        )

        # Projecting the energy of all sequences in the batch to a single sequence
        # of shape (SEQ_LEN, 1).
        a = self.v(energy).squeeze(-1)
        # Hence, our attention tensor 'a' should be of shape (BATCH_SIZE, SEQ_LEN).

        # Softmaxing along the SEQ_LEN dimension.
        return torch.softmax(a, dim=1)


class Decoder(nn.Module):

    def __init__(
        self, 
        vocab_size: int, 
        embed_size: int, 
        hidden_size: int, 
        padding_idx: int, 
        attention: nn.Module,
        dropout=0.1
    ):
        super().__init__()
        self.vocab_size = vocab_size
        self.attention = attention
        self.num_layers = 1

        self.embed = nn.Embedding(vocab_size, embed_size)
        self.gru = nn.GRU(
            input_size=(2 * hidden_size) + embed_size, 
            hidden_size=hidden_size,
            num_layers=self.num_layers,
            batch_first=True
        )
        self.dropout = nn.Dropout(dropout)

        self.project_out = nn.Linear(2 * hidden_size + hidden_size + embed_size, vocab_size)

    def forward(self, x: torch.Tensor, hidden: torch.Tensor, enc_out: torch.Tensor):
        # x.shape: (BATCH_SIZE, SEQ_LEN)
        # hidden.shape: (BATCH_SIZE, EMB_SIZE)
        # enc_out.shape: (BATCH_SIZE, DEC_EMB_SIZE * 2)

        x = x.unsqueeze(1)

        embedded_x = self.dropout(self.embed(x))
        # embedded_x.shape: (BATCH_SIZE, SEQ_LEN, EMB_SIZE)
        a = self.attention(hidden, enc_out)
        # Adding a 2nd dimension, since torch.bmm expects the two tensors to be
        # 3D.
        a = a.unsqueeze(1)
        # a.shape: (BATCH_SIZE, 1, SEQ_LEN)
        
        # The weighted sum of the encoder hidden states, using the attention
        # tensor.
        weighted = a @ enc_out
        # weighted.shape: (BATCH_SIZE, 1, (2 * ENC_HID_SIZE))

        gru_input = torch.cat((embedded_x, weighted), dim=-1)
        # gru_input.shape: (BATCH_SIZE, 1, (2 * ENC_HID_SIZE) + DEC_EMB_SIZE)

        # Since gru_input is batched (is 3D), hidden should also be batched, 
        # hence, we add an empty SEQ_LEN dimension.
        out, hidden = self.gru(gru_input, hidden.unsqueeze(0))
        # out.shape: (BATCH_SIZE, 1, DEC_HID_SIZE)
        # hidden.shape: (1, BATCH_SIZE, DEC_HID_SIZE)
        
        # Removing the SEQ_LEN dimension
        embedded_x = embedded_x.squeeze(1)
        out = out.squeeze(1)
        weighted = weighted.squeeze(1)

        pred = self.project_out(torch.cat((out, weighted, embedded_x), dim=-1))
        # pred.shape: (BATCH_SIZE, VOCAB_SIZE)

        return pred, hidden.squeeze(0)


class Seq2SeqSession:

    def __init__(
        self, 
        model: torch.nn.Module, 
        loss: torch.nn.Module, 
        optimizer: torch.optim.Optimizer, 
        vocab: Vocab, 
        device="cpu", 
        clip=5
    ):
        self.model = model
        self.loss_func = loss
        self.optimizer = optimizer

        self.vocab = vocab
        self.device = device
        self.clip = clip

        # Index to token/string.
        self.itos = {self.vocab[token]: token for token in self.vocab.vocab.itos_}
        # String/token to index.
        self.stoi = vocab

        self._init_weights(self.model)

    def train(
        self, 
        train_dataset: Dataset, valid_dataset: Dataset, 
        batch_size: int, epochs: int = 1, fixed_input: torch.LongTensor = None,
        num_workers: int = 0, teacher_forcing_ratio: float = 0.5, metrics: dict = None
    ):
        self.teacher_forcing_ratio = teacher_forcing_ratio

        train_dataloader = DataLoader(
            dataset=train_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True
        )
        valid_dataloader = DataLoader(
            dataset=valid_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True
        )
        self.fixed_input = fixed_input
        self.metrics = metrics

        for epoch in range(epochs):
            train_loss, train_metrics = self._train_epoch(train_dataloader)
            valid_loss, valid_metrics = self._valid_epoch(valid_dataloader)

            logging.info(f"Epoch: {epoch + 1}/{epochs}, Training Loss: {train_loss:.2f}, Validation Loss: {valid_loss:.2f}")
            self._log_metrics("Training", metrics=train_metrics)
            self._log_metrics("Validation", metrics=valid_metrics)
            logging.info("")

        return self.model

    def predict(self, word: torch.LongTensor):
        if len(word.shape) == 1:
            word = word.unsqueeze(0)

        seq_len = word.shape[-1]

        # Target with only a [START] token (and padding).
        target = [self.vocab["[START]"]] + [self.vocab["[PAD]"]] * (seq_len - 1)
        # Repeat the 'target' tensor along the row axis 'batch_size' times.
        target = torch.LongTensor(target).repeat(word.shape[0], 1).to(self.device)

        # Setting teacher_forcing_ratio to 0, since we want to use the prediction
        # generated by the model as the next input to the decoder.
        lemma = self.model(word, target, teacher_forcing_ratio=0)
        
        return lemma.argmax(-1)

    def _train_epoch(self, dataloader: DataLoader):
        for x, y in dataloader:
            x, y = x.to(self.device), y.to(self.device)

            y_pred = self.model(
                x, y, 
                teacher_forcing_ratio=self.teacher_forcing_ratio
            ).to(self.device)
            # y_pred's shape - (batch_size, seq_len, vocab_size)
            
            # Reshaping y_pred:
            vocab_size = y_pred.shape[-1]
            y_pred = y_pred[:, 1:, :].reshape(-1, vocab_size)

            # Reshaping y:
            y = y[:, 1:].reshape(-1)
            
            self.optimizer.zero_grad()

            loss = self.loss_func(y_pred, y)
            loss.backward()

            # Clipping the gradients, since LSTMs/GRUs have Exploding Gradient problems.
            nn.utils.clip_grad_norm_(self.model.parameters(), self.clip)
            self.optimizer.step()

        metrics = None
        if self.metrics is not None:
            metrics = self._get_metrics_results(y_pred, y)

        return loss.item(), metrics

    def _valid_epoch(self, dataloader: DataLoader):
        self.model.eval()

        with torch.no_grad():
            for x, y in dataloader:
                x, y = x.to(self.device), y.to(self.device)
                
                y_pred = self.model(x, y, teacher_forcing_ratio=0).to(self.device)

                vocab_size = y_pred.shape[-1]
                y_pred = y_pred[:, 1:, :].reshape(-1, vocab_size)

                y = y[:, 1:].reshape(-1)

                loss = self.loss_func(y_pred, y)

            if self.fixed_input is not None:
                fixed_pred = self.predict(word=self.fixed_input)

                fixed_input = self.fixed_input.tolist()
                fixed_inputs = []
                for seq in fixed_input:
                    fixed_inputs.append("".join([self.itos[int(idx)] for idx in seq]))

                fixed_preds = []
                for seq in fixed_pred:
                    fixed_preds.append("".join([self.itos[int(idx)] for idx in seq]))

                fixed_input = '\n'.join(fixed_inputs)
                fixed_pred = '\n'.join(fixed_preds)

                logging.info(f"Inputs:\n{fixed_input}")
                logging.info(f"Predictions:\n{fixed_pred}")

        metrics = None
        if self.metrics is not None:
            metrics = self._get_metrics_results(y_pred, y)

        self.model.train()

        return loss.item(), metrics

    def _init_weights(self, m: nn.Module):
        # This initialization comes from the Seq2Seq paper.
        for name, param in m.named_parameters():
            nn.init.uniform_(param.data, -0.08, 0.08)

    def _get_metrics_results(self, y_pred: torch.Tensor, y: torch.Tensor):
        results = {}

        for metric_name, metric in self.metrics.items():
            y_pred_argmax = y_pred.argmax(-1)
            result = metric(y_pred_argmax, y)

            results[metric_name] = float(result)

        return results

    def _log_metrics(self, type_: str, metrics: dict):
        log_messages = []
        for metric_name, metric in metrics.items():
            log_messages.append(f"{type_} {metric_name}: {metric:.2f}")

        logging.info(", ".join(log_messages))



from typing import List


class SubwordMixin:

    def subwords_to_words(self, tokens: List[str]) -> List[str]:
        """Covert subword tokens into word tokens.
        
        Example::
            >>> subword_tokens = ["▁Здра", "вей", "▁другарю", "!"]
            >>> tokens = SubwordMixin().subwords_to_words(subword_tokens)
            >>> tokens
            ["Здравей", "другарю", "!"]
        """
        out_tokens = []
        curr_token = ""
        
        for token in tokens:
            if token == "[SEP]":
                curr_token = curr_token.replace("▁", "")
                out_tokens.append(curr_token)
                out_tokens.append("[SEP]")
                break

            if "▁" in token and curr_token == "":
                curr_token += token

            elif "▁" in token and curr_token != "":
                curr_token = curr_token.replace("▁", "")
                out_tokens.append(curr_token)
                curr_token = ""
                curr_token += token

            elif "▁" not in token:
                curr_token += token

        return out_tokens


"""
Natural language Taggers for Bulgarian. They are all model-based.
"""
import os
import re
from typing import List, Dict
from abc import ABC, abstractmethod

import torch
from torch import nn
import gdown
from transformers import logging
from transformers import AutoTokenizer, AutoModelForTokenClassification

from bgnlp.models import LemmaBert
from bgnlp.tools.mixins import SubwordMixin
from bgnlp.tools.tokenizers import (
    CharacterBasedPreTokenizer, CharacterBasedTokenizer
)
from bgnlp.tools.configs import LemmaTaggerConfig, ModelConfig, PosTaggerConfig


# Logging only error messages from HuggingFace.
logging.set_verbosity_error()


class BaseTagger(ABC):

    @abstractmethod
    def get_tokenizer(self):
        pass

    @abstractmethod
    def get_model(self):
        pass

    @abstractmethod
    def predict(self):
        pass

    def load_model(self, model_obj):
        if os.path.exists(self.config.model_path):
            model_obj.load_state_dict(torch.load(self.config.model_path, map_location=self.config.device))
        else:
            # Downloading the model if it doesn't exist locally.
            # The model is not deployed with the PyPI package - hence, the download below.
            gdown.download(
                self.config.model_url, 
                self.config.model_path, 
                quiet=False
            )
            model_obj.load_state_dict(torch.load(self.config.model_path, map_location=self.config.device))

        return model_obj


class PosTagger(BaseTagger, SubwordMixin):
    """Part-of-speech tagger. Tagging is done using a BERT model trained on 
    [Wiki1000+ Bulgarian corpus](http://dcl.bas.bg/wikiCorpus.html).
    
    Args:
        config (ModelConfig): Configuration of the PosTagger.
    """

    def __init__(self, config: ModelConfig):
        self.config = config
        self.tokenizer = self.get_tokenizer()
        self.model = self.get_model().to(self.config.device)

        # TODO: Have better descriptions. Figure out the full meanings of the tags.
        self.TAGS_MAPPING = {
            "N": {
                "en": "noun",
                "bg": "съществително име"
            },
            "A": {
                "en": "adjective",
                "bg": "прилагателно име"
            },
            "P": {
                "en": "pronoun",
                "bg": "местоимение"
            },
            "B": {
                "en": "numeral",
                "bg": "числително име"
            },
            "V": {
                "en": "verb",
                "bg": "глагол"
            },
            "D": {
                "en": "adverb",
                "bg": "наречие"
            },
            "C": {
                "en": "conjunction",
                "bg": "съюз"
            },
            "T": {
                "en": "particle",
                "bg": "частица"
            },
            "R": {
                "en": "preposition",
                "bg": "предлог"
            },
            "I": {
                "en": "interjection",
                "bg": "междуметие"
            },
            "U": {
                "en": "punctuation",
                "bg": "препинателен знак"
            }
        }

    def __call__(self, text: str, max_len=64) -> List[Dict[str, str]]:
        """Tag each one of the words in `text` with a part-of-speech tag.

        Args:
            text (str): Text in Bulgarian.
            max_len (int, optional): The maximum number of words that you can have in `text`. Defaults to 64.

        Returns:
            List[Dict[str, str]]: List of dictionaries for each word and its tags.

        Example::

            >>> from bgnlp import PosTagger, PosTaggerConfig
            >>> 
            >>> 
            >>> config = PosTaggerConfig()
            >>> pos = PosTagger(config=config)
            >>> pos("Това е библиотека за обработка на естествен език.")
            [{
                "word": "Това",
                "tag": "PDOsn",
                "bg_desc": "местоимение",
                "en_desc": "pronoun"
            }, {
                "word": "е",
                "tag": "VLINr3s",
                "bg_desc": "глагол",
                "en_desc": "verb"
            }, {
                "word": "библиотека",
                "tag": "NCFsof",
                "bg_desc": "съществително име",
                "en_desc": "noun"
            }, {
                "word": "за",
                "tag": "R",
                "bg_desc": "предлог",
                "en_desc": "preposition"
            }, {
                "word": "обработка",
                "tag": "NCFsof",
                "bg_desc": "съществително име",
                "en_desc": "noun"
            }, {
                "word": "на",
                "tag": "R",
                "bg_desc": "предлог",
                "en_desc": "preposition"
            }, {
                "word": "естествен",
                "tag": "Asmo",
                "bg_desc": "прилагателно име",
                "en_desc": "adjective"
            }, {
                "word": "език",
                "tag": "NCMsom",
                "bg_desc": "съществително име",
                "en_desc": "noun"
            }, {
                "word": ".",
                "tag": "U",
                "bg_desc": "препинателен знак",
                "en_desc": "punctuation"
            }]
        """
        self.max_len = max_len

        return self.predict(text)

    def predict(self, text: str) -> List[Dict[str, str]]:
        """Tag each one of the words in `text` with a part-of-speech tag.

        Args:
            text (str): Text in Bulgarian.

        Returns:
            List[Dict[str, str]]: List of dictionaries for each word and its tags.
        """
        text = PosTagger._preprocess_text(text)

        tokens_data = self.tokenizer(
            text,
            max_length=self.max_len,
            padding="max_length"
        )

        # Preparing the input to the model.
        input_ids, attn = tokens_data["input_ids"], tokens_data["attention_mask"]
        input_ids = torch.LongTensor(input_ids).unsqueeze(0).to(self.config.device)
        attn =  torch.LongTensor(attn).unsqueeze(0).to(self.config.device)

        # Making a prediction.
        self.model.eval()
        with torch.no_grad():
            pred = self.model(input_ids, attention_mask=attn).logits
            pred = pred.argmax(-1).squeeze(0)
        self.model.train()

        pred = [
            self.config.id2label[int(id_)] 
            for id_ in pred
        ]

        return self._format_prediction(
            input_tokens=self.tokenizer.convert_ids_to_tokens(
                tokens_data["input_ids"]
            ), 
            prediction=pred
        )

    def get_tokenizer(self) -> AutoTokenizer:
        """Get the tokenizer for the used model.

        Returns:
            AutoTokenizer: HuggingFace AutoTokenizer.
        """
        return AutoTokenizer.from_pretrained(self.config.base_model_id)

    def get_model(self) -> nn.Module:
        """Get the model used for tagging. When this method is called for the first
        time, the model is downloaded. Afterwards, it should be part of your package.

        Returns:
            nn.Module: PyTorch Module.
        """
        bert = AutoModelForTokenClassification.from_pretrained(
            self.config.base_model_id, 
            num_labels=len(self.config.label2id), 
            label2id=self.config.label2id
        )
        bert.resize_token_embeddings(len(self.tokenizer))
        bert = self.load_model(bert)

        return bert

    @staticmethod
    def _preprocess_text(text: str):
        """Prepare the string `text` for the model. 
        
        This includes:
        - Surrounding punctuation with whitespace
        - Converting multiple consecutive whitespaces into one.

        Args:
            text (str): _description_

        Returns:
            str: _description_
        """
        text = re.sub(r"([.,!?:;])+", r" \1 ", text)
        text = re.sub(r"(\s+)", " ", text)

        return text.strip()
    
    def _format_prediction(self, input_tokens: List[str], prediction: List[str]) -> List[Dict[str, str]]:
        """Format the prediction returned from the model.

        Since the tokenizer of the model is a subword tokenizer, the words
        are split into multiple subwords. The task of this method is to merge 
        them and then create a dictionaries with their tags.

        Args:
            input_tokens (List[str]): Input words (tokens) as strings.
            prediction (List[str]): Predicted tags.

        Returns:
            List[Dict[str, str]]: All found words and their tags.
        """
        tags = []

        # This method is taken from the SubwordMixin class.
        # Here, I am passing all tokens except for the 1st one - [CLS].
        tokens = self.subwords_to_words(input_tokens[1:])

        # Getting all predicted tokens (except [CLS]) up until [SEP].
        for token in prediction[1:]:
            if token == "[SEP]":
                break
            tags.append(token)

        result = []
        for word, tag in zip(tokens, tags):
            result.append({
                "word": word,
                "tag": tag,
                "bg_desc": self._get_tag_description(lang="bg", tag=tag),
                "en_desc": self._get_tag_description(lang="en", tag=tag)
            })

        return result

    def _get_tag_description(self, lang: str, tag: str) -> str:
        """Map `tag` to its description based on `lang`

        Args:
            lang (str): Language - either 'bg' or 'en'.
            tag (str): PoS tag as a string.

        Returns:
            str: The hardcoded tag description, based on `self.TAGS_MAPPING`.
        """
        first_tag = tag[0]
        description = self.TAGS_MAPPING[first_tag][lang]

        return description


class Lemmatizer(BaseTagger):
    """Lemmatize a word. This is only for single-word lemmatization. If you want 
    to lemmatize multiple words, please use :ref:`LemmaTagger`.

    Args:
        config (ModelConfig): Configuration of the :ref:`LemmaBert` model.
    """

    def __init__(self, config: ModelConfig):
        self.config = config
        self.tokenizer = self.get_tokenizer()
        self.model = self.get_model()

    def __call__(self, word: str, pos: str) -> str:
        """Convert `word` into its lemma.

        Args:
            word (str): Word in Bulgarian.
            pos (str): Its part-of-speech tag.

        Returns:
            str: The lemma of `word`.

        Example::
            >>> from bgnlp import LemmaTaggerConfig
            >>> from bgnlp.tools.taggers import Lemmatizer
            >>> 
            >>> 
            >>> lemma = Lemmatizer(config=LemmaTaggerConfig())
            >>> lemma("езикът", "Ns")
            език
        """
        return self.predict(word, pos)

    def get_tokenizer(self) -> CharacterBasedTokenizer:
        """Get the tokenizer used by `LemmaBert`. It is a character-based one.

        Returns:
            CharacterBasedTokenizer: The tokenizer.
        """
        vocab = torch.load(self.config.vocab_path)
        pretokenizer = CharacterBasedPreTokenizer()

        tokenizer = CharacterBasedTokenizer(
            pretokenizer=pretokenizer,
            vocab=vocab
        )

        return tokenizer

    def get_model(self) -> nn.Module:
        """Get the `LemmaBert` model.

        Returns:
            nn.Module: PyTorch Module.
        """
        bert = LemmaBert(
            vocab_size=len(self.tokenizer), 
            output_size=len(self.tokenizer),
            device=self.config.device
        ).to(self.config.device)
        bert = self.load_model(bert)

        return bert

    def predict(self, word: str, pos: str) -> str:
        """Convert `word` into its lemma.

        Args:
            word (str): Word in Bulgarian.
            pos (str): Its part-of-speech tag.

        Returns:
            str: The lemma of `word`.
        """
        # Preparing the input.
        tokens, attention_mask = self.tokenizer(word, pos)

        tokens = torch.LongTensor(tokens).unsqueeze(0)
        attention_mask = torch.LongTensor(attention_mask).unsqueeze(0)

        self.model.eval()

        with torch.no_grad():
            pred = self.model(tokens, attention_mask=attention_mask).argmax(-1).squeeze(0).tolist()
            pred = "".join(self.tokenizer.vocab.lookup_tokens(pred))
            pred = re.findall(r"\[CLS\](.+?)\[SEP\]", pred)[0]

        self.model.train()

        if word[0].isupper():
            pred = pred.capitalize()

        return pred


class LemmaTagger:
    """Find the lemmas of a string with one or more words.

    Args:
        config (ModelConfig): Configuration of the :ref:`LemmaBert` model.
    """

    def __init__(self, config: ModelConfig):
        self.config = config

    def __call__(self, text: str, additional_info: bool = False) -> List[Dict[str, str]]:
        """Find the lemmas of `text`. `text` should preferably be a semantically correct sentence or sentences, since the lemma sometimes changes based on context.

        Args:
            text (str): String with one or more Bulgarian words.
            additional_info (bool, optional): Whether the output should constist of more data about each word (mainly PoS information). Defaults to False.

        Returns:
            List[Dict[str, str]]: List of dictionaries. Each dictionary has a word and a `lemma` key with a value - its lemma. If `additional_info`=True, the dictionary has PoS data.

        Example::
            >>> from bgnlp import LemmaTaggerConfig, LemmaTagger
            >>> 
            >>> 
            >>> lemma = LemmaTagger(config=LemmaTaggerConfig())
            >>> text = "Добре дошли!"
            >>> print("Input:", text)
            >>> print("Output:", lemma(text))
        """
        pos = PosTagger(config=PosTaggerConfig())
        lemma = Lemmatizer(config=LemmaTaggerConfig)
        result = []

        for pos_result in pos(text):
            pos_result["lemma"] = lemma(
                word=pos_result["word"],
                pos=pos_result["tag"]
            )
            if additional_info:
                result.append(pos_result)
            else:
                result.append({
                    "word": pos_result["word"],
                    "lemma": pos_result["lemma"]
                })

        return result


class NerTagger(BaseTagger, SubwordMixin):

    def __init__(self, config: ModelConfig):
        self.config = config
        self.tokenizer = self.get_tokenizer()
        self.model = self.get_model()

    def __call__(self, text: str):
        return self.predict(text)

    def get_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.config.model_path)

    def get_model(self):
        return AutoModelForTokenClassification.from_pretrained(self.config.model_path)

    def predict(
        self, 
        text, 
        label2id={
            0: "O",
            1: "B-PER", 2: "I-PER", 
            3: "B-ORG", 4: "I-ORG", 
            5: "B-LOC", 6: "I-LOC"
        }
    ):
        tokens_data = self.tokenizer(text)
        tokens = self.tokenizer.convert_ids_to_tokens(tokens_data["input_ids"])
        words = self.subwords_to_words(tokens)

        input_ids = torch.LongTensor(tokens_data["input_ids"]).unsqueeze(0)
        attention_mask = torch.LongTensor(tokens_data["attention_mask"]).unsqueeze(0)

        out = self.model(input_ids, attention_mask=attention_mask).logits
        out = out.argmax(-1).squeeze(0).tolist()

        prediction = [label2id[idx] if idx in label2id else idx for idx in out]

        return self._merge_words_and_predictions(
            words=words, entities=prediction
        )
    
    def _merge_words_and_predictions(self, words: List[str], entities: List[str]) -> List[Dict[str, str]]:
        result = []
        curr_word = []

        for i, (word, entity) in enumerate(zip(words[1:], entities[1:])):
            if "B-" in entity:
                if curr_word:
                    curr_word = " ".join(curr_word)
                    result.append({
                        "word": curr_word,
                        "entity_group": entities[i][2:]
                    })
                    curr_word = [word]
                else:
                    curr_word.append(word)

            if "I-" in entity:
                curr_word.append(word)
            
            if "O" == entity:
                if curr_word:
                    curr_word = " ".join(curr_word)
                    result.append({
                        "word": curr_word,
                        "entity_group": entities[i][2:]
                    })
                
                curr_word = []

        return result



import re

from torchtext.vocab import Vocab


class CharacterBasedPreTokenizer:
    
    def __call__(self, text: str):
        text = re.sub(r"([,\.\/!?;:\'\"])", r" \1 ", text)
        text = re.sub(r"\s+", " ", text)
        text = text.strip()
        
        return list(text)


class CharacterBasedTokenizer:
    
    def __init__(self, pretokenizer: CharacterBasedPreTokenizer, vocab: Vocab):
        self.pretokenizer = pretokenizer
        # Mapping the tokens to IDs using vocab.
        # Inversely, we can use self.vocab.lookup_token(8) to map
        # the IDs to their respective tokens.
        self.vocab = vocab
        
    def __call__(
        self, 
        word, pos, 
        index=True, pad=True, truncate=False, max_length=50,
        cls_token="[CLS]", sep_token="[SEP]", pad_token="[PAD]"
    ):
        tokens = [cls_token] + self.pretokenizer(word) + [sep_token] + [pos] + [sep_token]

        if pad:
            if max_length > len(tokens):
                padding = [pad_token] * (max_length - len(tokens))
                tokens += padding
            
            if truncate:
                if max_length <= len(tokens):
                    tokens = tokens[:max_length]
        
        # Adding attention mask, which masks the padding.
        attention_mask = [1 if token != pad_token else 0 for token in tokens]

        if index:
            return [self.vocab[token] for token in tokens], attention_mask

        return tokens, attention_mask
    
    def __len__(self):
        return len(self.vocab)



import os
import re
from pathlib import Path
from setuptools import setup, find_packages


ROOT = os.path.abspath(os.path.dirname(__file__))
README_PATH = os.path.join(ROOT, "README.md")
REQUIREMENTS_PATH = os.path.join(ROOT, "requirements.txt")

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

VERSION = '0.0.12'
DESCRIPTION = 'Package for Bulgarian Natural Language Processing (NLP)'


def _get_requirements(path):
    with open(path, "r") as f:
        requirements_str = f.read()
        packages = re.findall(r"(.+==[^\n]+)\n", requirements_str)
        return packages


if __name__ == "__main__":
    setup(
        name="bgnlp",
        version=VERSION,
        author="Adam Fauzi",
        author_email="adamfzh98@gmail.com",
        description=DESCRIPTION,
        long_description_content_type="text/markdown",
        long_description=long_description,
        packages=find_packages(),
        install_requires=_get_requirements(REQUIREMENTS_PATH),
        keywords=['pytorch', 'nlp', 'bulgaria', 'machine learning', "deep learning", "AI"],
        classifiers=[
            "Development Status :: 1 - Planning",
            "Intended Audience :: Developers",

            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Mathematics',
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',

            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Operating System :: Unix",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
        ],
        # This is used because we need the resource files of the package (mainly models).
        include_package_data=True
    )
